
import UIKit
import CloudKit

enum Sections {
   case main
}
class ApplicationData {
   var database: CKDatabase!
   var selectedCountry: CKRecord.ID!

   var dataSourceCountries: UITableViewDiffableDataSource<Sections, CKRecord.ID>!
   var listCountries: [CKRecord] = []

   var dataSourceCities: UITableViewDiffableDataSource<Sections, CKRecord.ID>!
   var listCities: [CKRecord] = []

   init() {
      let container = CKContainer.default()
      database = container.privateCloudDatabase
   }
   
   func configureDatabase() async {
      let userSettings = UserDefaults.standard
      if !userSettings.bool(forKey: "subscriptionSaved") {
         let newSubscription = CKDatabaseSubscription(subscriptionID: "updatesDatabase")
         let info = CKSubscription.NotificationInfo()
         info.shouldSendContentAvailable = true
         newSubscription.notificationInfo = info

         do {
            try await database.save(newSubscription)
            userSettings.set(true, forKey: "subscriptionSaved")
         } catch {
            print("Error: \(error)")
         }
      }
      if !userSettings.bool(forKey: "zoneCreated") {
         let newZone = CKRecordZone(zoneName: "listPlaces")
         do {
            try await database.save(newZone)
            userSettings.set(true, forKey: "zoneCreated")
         } catch {
            print("Error: \(error)")
         }
      }
   }

   func checkUpdates(finishClosure: @escaping (UIBackgroundFetchResult) -> Void) {
      Task(priority: .background) {
         await configureDatabase()
         downloadUpdates(finishClosure: finishClosure)
      }
   }

   func downloadUpdates(finishClosure: @escaping (UIBackgroundFetchResult) -> Void) {
      var changeToken: CKServerChangeToken!
      var changeZoneToken: CKServerChangeToken!

      let userSettings = UserDefaults.standard
      if let data = userSettings.value(forKey: "changeToken") as? Data {
         if let token = try? NSKeyedUnarchiver.unarchivedObject(ofClass: CKServerChangeToken.self, from: data) {
            changeToken = token
         }
      }
      if let data = userSettings.value(forKey: "changeZoneToken") as? Data {
         if let token = try? NSKeyedUnarchiver.unarchivedObject(ofClass: CKServerChangeToken.self, from: data) {
            changeZoneToken = token
         }
      }
      var zonesIDs: [CKRecordZone.ID] = []
      let operation = CKFetchDatabaseChangesOperation(previousServerChangeToken: changeToken)
      operation.recordZoneWithIDChangedBlock = { zoneID in
         zonesIDs.append(zoneID)
      }
      operation.changeTokenUpdatedBlock = { token in
         changeToken = token
      }
      operation.fetchDatabaseChangesResultBlock = { result in
         guard let values = try? result.get() else {
            finishClosure(UIBackgroundFetchResult.failed)
            return
         }
         if zonesIDs.isEmpty {
            finishClosure(UIBackgroundFetchResult.noData)
         } else {
            changeToken = values.serverChangeToken

            let configuration = CKFetchRecordZoneChangesOperation.ZoneConfiguration()
            configuration.previousServerChangeToken = changeZoneToken
            let fetchOperation = CKFetchRecordZoneChangesOperation(recordZoneIDs: zonesIDs, configurationsByRecordZoneID: [zonesIDs[0]: configuration])

            fetchOperation.recordWasChangedBlock = { recordID, result in
               guard let record = try? result.get() else {
                  print("Error")
                  return
               }
               if record.recordType == "Countries" {
                  let index = self.listCountries.firstIndex(where: { item in
                     return item.recordID == recordID
                  })
                  if index != nil {
                     self.listCountries[index!] = record
                  } else {
                     self.listCountries.append(record)
                  }
               } else if record.recordType == "Cities" {
                  if let country = record["country"] as? CKRecord.Reference {
                     if country.recordID == self.selectedCountry {
                        let index = self.listCities.firstIndex(where: { item in
                           return item.recordID == record.recordID
                        })
                        if index != nil {
                           self.listCities[index!] = record
                        } else {
                           self.listCities.append(record)
                        }
                     }
                  }
               }
            }
            fetchOperation.recordWithIDWasDeletedBlock = { recordID, recordType in
               if recordType == "Countries" {
                  let index = self.listCountries.firstIndex(where: { item in
                     return item.recordID == recordID
                  })
                  if index != nil {
                     self.listCountries.remove(at: index!)
                  }
               } else if recordType == "Cities" {
                  let index = self.listCities.firstIndex(where: { item in
                     return item.recordID == recordID
                  })
                  if index != nil {
                     self.listCities.remove(at: index!)
                  }
               }
            }
            fetchOperation.recordZoneChangeTokensUpdatedBlock = { zoneID, token, data in
               changeZoneToken = token
            }
            fetchOperation.recordZoneFetchResultBlock = { zoneID, result in
               guard let values = try? result.get() else {
                  print("Error")
                  return
               }
               changeZoneToken = values.serverChangeToken
            }
            fetchOperation.fetchRecordZoneChangesResultBlock = { result in
               switch result {
                  case .failure(_):
                     finishClosure(UIBackgroundFetchResult.failed)
                     return
                  default:
                     break
               }
               if changeToken != nil {
                  if let data = try? NSKeyedArchiver.archivedData(withRootObject: changeToken!, requiringSecureCoding: false) {
                     userSettings.set(data, forKey: "changeToken")
                  }
               }
               if changeZoneToken != nil {
                  if let data = try? NSKeyedArchiver.archivedData(withRootObject: changeZoneToken!, requiringSecureCoding: false) {
                     userSettings.set(data, forKey: "changeZoneToken")
                  }
               }
               Task(priority: .background) {
                  await self.updateInterface()
               }
               finishClosure(UIBackgroundFetchResult.newData)
            }
            self.database.add(fetchOperation)
         }
      }
      database.add(operation)
   }
   func updateInterface() async {
      await MainActor.run {
         var snapshotCountries = NSDiffableDataSourceSnapshot<Sections, CKRecord.ID>()
         snapshotCountries.appendSections([.main])
         snapshotCountries.appendItems(listCountries.map({ $0.recordID }))
         dataSourceCountries?.apply(snapshotCountries)

         var snapshotCities = NSDiffableDataSourceSnapshot<Sections, CKRecord.ID>()
         snapshotCities.appendSections([.main])
         snapshotCities.appendItems(listCities.map({ $0.recordID }))
         dataSourceCities?.apply(snapshotCities)
      }
   }

}
var AppData = ApplicationData()

